import 'dart:mirrors';

import 'package:nats_dart/nats_dart.dart';
import 'package:test/test.dart';
import 'package:faker_extended/faker_extended.dart';

void main() {
  var testCases = [
    new testCase('BAD_AUTHENTICATION'                                                  ),
    new testCase('BAD_CREDENTIALS',                  aliases: ['BAD_CREDS'    ]        ),
    new testCase('BAD_NKEY_CREDENTIALS',             aliases: ['BAD_NKEY_SEED']        ),
    new testCase('BAD_JSON'                                                            ),
    new testCase('BAD_MESSAGE',                      aliases: ['BAD_MSG']              ),
    new testCase('BAD_REPLY'                                                           ),
    new testCase('BAD_SUBJECT'                                                         ),
    new testCase('CLIENT_CERT_REQUIRED',             aliases: ['CLIENT_CERT_REQ']      ),
    new testCase('CONNECTION_CLOSED',                aliases: ['CONN_CLOSED'    ]      ),
    new testCase('CONNECTION_DRAINING',              aliases: ['CONN_DRAINING'  ]      ),
    new testCase('CONNECTION_ERROR',                 aliases: ['CONN_ERR'       ]      ),
    new testCase('CONNECTION_TIMEOUT',               aliases: ['CONN_TIMEOUT'   ]      ),
    new testCase('INVALID_ENCODING'                                                    ),
    new testCase('NKEY_OR_JWT_REQUIRED',             aliases: ['NKEY_OR_JWT_REQ']      ),
    new testCase('NO_ECHO_NOT_SUPPORTED'                                               ),
    new testCase('NO_SEED_IN_CREDENTIALS',           aliases: ['NO_SEED_IN_CREDS'     ]),
    new testCase('NO_USER_JWT_IN_CREDENTIALS',       aliases: ['NO_USER_JWT_IN_CREDS' ]),
    new testCase('NON_SECURE_CONNECTION_REQUIRED',   aliases: ['NON_SECURE_CONN_REQ'  ]),
    new testCase('NONCE_SIGNER_NOT_FUNCTION',        aliases: ['NONCE_SIGNER_NOT_FUNC']),
    new testCase('REQUEST_TIMEOUT',                  aliases: ['REQ_TIMEOUT'          ]),
    new testCase('SECURE_CONNECTION_REQUIRED',       aliases: ['SECURE_CONN_REQ'      ]),
    new testCase('SIGNATURE_REQUIRED',               aliases: ['SIG_REQ'              ]),
    new testCase('SUBSCRIPTION_CLOSED',              aliases: ['SUB_CLOSED'           ]),
    new testCase('SUBSCRIPTION_DRAINING',            aliases: ['SUB_DRAINING'         ]),
    new testCase('SUBSCRIPTION_TIMEOUT',             aliases: ['SUB_TIMEOUT'          ]),
    new testCase('UNKNOWN'                                                             ),
  ];

  /// string names/aliases are registered with the static "constants" are first referenced. If these constants have never been referenced, getting/initializing these by string should still work.
  /// NOTE: this group should always run first to ensure the string names/aliases are resolving EVEN IF the statics haven't yet been referenced
  group('accessible by string', () {
    group('(name)', () {
      testCases.forEach((testCase) => test(testCase.name, () {
        var actual = NatsError.fromString(testCase.name);
        expect(actual.code, same(testCase.name), reason: "Expected enumeration name(string) not defined");
        testCase.expected = actual; // -- add the actual reference to our testCases array
      }));
    });

    group('(alias)', () => testCases.forEach((testCase) {
      group(testCase.name, () {
        testCase.aliases.forEach((alias) {
          test(alias, () => expect(NatsError.fromString(alias), same(testCase.expected), reason: "Expected enumeration name(string) not defined"));
        });
      });
    }));
  });

  group('constants exist', () {
    ClassMirror reflection = reflectClass(NatsError);

    testCases.map((testCase) => testCase.name).forEach((testCase) => test(testCase, () {
      expect(() => reflection.getField(Symbol(testCase)), returnsNormally, reason: "Expected enumeration value not defined");
    }));
  });

  group('can be nested', () {
    group('from instance', () {
      NatsError base;
      setUp(() {
        base = faker.randomGenerator.element(testCases.map<NatsError>((testCase) => testCase.expected).toList());
      });

      test('caused by NatsError', () {
        var cause = faker.randomGenerator.element(testCases.map<NatsError>((testCase) => testCase.expected).toList());

        var actual = NatsError.from(base, cause: cause);
        expect(actual,       equals(base),      reason: "returned instance is of an unexpected type"                                                               );
        expect(actual,       isNot(same(base)), reason: "returned instance is unexpectedly the exact same instance as the base type rather than an extension of it");
        expect(actual.cause, same(cause),       reason: "returned instance is unexpectedly not associated with the given cause"                                    );
      });

      test('caused by Error', () {
        var cause  = new AssertionError();
        expect(() => NatsError.from(base, cause: cause), returnsNormally);
      });

      test('caused by Exception', () {
        var cause  = new FormatException();
        expect(() => NatsError.from(base, cause: cause), returnsNormally);
      });

      test('caused by other', () {
        var cause  = faker.lorem.word();
        expect(() => NatsError.from(base, cause: cause), throwsA(TypeMatcher<AssertionError>()));
      });
    });

    group('from string', () {
      group('(by name)', () {
        NatsError expectedBase;
        String    baseName;

        setUp(() {
          expectedBase = faker.randomGenerator.element(testCases.map<NatsError>((testCase) => testCase.expected).toList());
          baseName     = expectedBase.code;
        });

        test('caused by NatsError', () {
          NatsError cause  = faker.randomGenerator.element(testCases.map<NatsError>((testCase) => testCase.expected).toList());
          NatsError actual = NatsError.fromString(baseName, cause: cause);

          expect(actual,       equals(expectedBase),      reason: "returned instance is of an unexpected type"                                                               );
          expect(actual,       isNot(same(expectedBase)), reason: "returned instance is unexpectedly the exact same instance as the base type rather than an extension of it");
          expect(actual.cause, same(cause),               reason: "returned instance is unexpectedly not associated with the given cause"                                    );
        });

        test('caused by Error', () {
          var cause  = new AssertionError();
          expect(() => NatsError.fromString(baseName, cause: cause), returnsNormally);
        });

        test('caused by Exception', () {
          var cause  = new FormatException();
          expect(() => NatsError.fromString(baseName, cause: cause), returnsNormally);
        });

        test('caused by other', () {
          var cause  = faker.lorem.word();
          expect(() => NatsError.fromString(baseName, cause: cause), throwsA(TypeMatcher<AssertionError>()));
        });
      });

      group('(by alias)', () {
        final List aliasedTestCases = testCases.where((testCase) => testCase.aliases.isNotEmpty).toList();

        String baseAlias;
        Object expectedBase;

        setUp(() {
          var randomTestCase = faker.randomGenerator.element(aliasedTestCases);
          baseAlias          = faker.randomGenerator.element(randomTestCase.aliases);
          expectedBase       = randomTestCase.expected;
        });

        test('caused by NatsError', () {
          NatsError cause    = faker.randomGenerator.element(testCases.map<NatsError>((testCase) => testCase.expected).toList());
          NatsError actual   = NatsError.fromString(baseAlias, cause: cause);

          expect(actual,       equals(expectedBase),      reason: "returned instance is of an unexpected type"                                                               );
          expect(actual,       isNot(same(expectedBase)), reason: "returned instance is unexpectedly the exact same instance as the base type rather than an extension of it");
          expect(actual.cause, same(cause),               reason: "returned instance is unexpectedly not associated with the given cause"                                    );
        });

        test('caused by Error', () {
          var cause  = new AssertionError();
          expect(() => NatsError.fromString(baseAlias, cause: cause), returnsNormally);
        });

        test('caused by Exception', () {
          var cause  = new FormatException();
          expect(() => NatsError.fromString(baseAlias, cause: cause), returnsNormally);
        });

        test('caused by other', () {
          var cause  = faker.lorem.word();
          expect(() => NatsError.fromString(baseAlias, cause: cause), throwsA(TypeMatcher<AssertionError>()));
        });
      });
    });
  });
}

class testCase {
  String       name;
  List<String> aliases;
  Object       expected;

  testCase(this.name, {this.aliases = const [], this.expected});
}
